#!/usr/bin/env python
# encoding: UTF-8
# api: python
# type: application
# title: streamtuner2
# description: directory browser for internet radio / audio streams
# depends: gtk, pygtk, xml.dom.minidom, threading
# version: 1.9.8
# author: mario
# license: public domain
# url: http://freshmeat.net/p/streamtuner2
#
#
# Streamtuner2 is a GUI browser for internet radio directories. Various
# providers can be added, and streaming stations are usually grouped into
# music genres or categories.
#
# It's an independent rewrite of streamtuner1 in a scripting language. So
# it can be more easily extended and fixed. It still lacks many features
# and some might get left out even.
# Therefore streamtuner2 should be considered just an alternative to the
# original. It's not necessarily a replacement for everyone.
#
# As it is in a very early stage (just four weeks of development), it also
# isn't much more stable. It works rudimentarily, the GUI is very wacky
# however, and it has neither been optimized for memory use or performance
# yet. The internal workings are supposed to develop further... (not best
# structure right now.)  And documentation is lacking.
#


#  features:
#   - less functionality than original C streamtuner 0.99.x
#   - treeview lists are created from datamap[] structure and stream{} dicts
#   - channel categories are built-in defaults (can be freshened up however)
#   - config vars and cache data get stored as JSON in ~/.config/streamtuner2/
#
#  missing:
#   - favicons
#   - right click context menus
#   - localization
#   - per-channel .play() method instead of using url hints
#   - search feature
#   - unicode use
#
#  security notes:
#   - directory scrapers use fragile regular expressions - which is probably
#     not a security risk, but might lead to faulty data
#   - MEDIUM: little integrity checking for .pls / .m3u references and files
#   - minimal XML/SGML entity decoding (-> faulty data)
#   - MINIMAL: if system json module is not available, pseudo-json uses eval()
#     to read the config data -> limited risk, since it's only local files
#   - HIGH RISK: no verification of downloaded favicon image files (ico/png),
#     as they are passed to gtk.gdk.Pixbuf <= not implemented yet, anyhow
#   - MEDIUM: audio players / decoders are prone to buffer overflows, and
#     streamtuner2 executes them - well, that's the purpose -> no workaround
#
#  help needed on:
#   - any of the above
#   - logo
#   - more professional python app structure
#
#  current todo
#  - plugin api ST for st1 scripts
#  - favicons
#



# standard modules
import sys
import os
import re
import xml.dom.minidom
from xml.sax.saxutils import unescape as entity_decode, escape as xmlentities
import copy
import urllib

# gtk modules
import pygtk
import gtk
import gtk.glade
import gobject

# custom modules
sys.path.append("/usr/share/streamtuner2")
import json   # eventually loads emulation
from processing import Process as Thread  # loads emulation with threading. on Python <= 2.5
from config import conf
from mygtk import mygtk
import http
import action  # needs workaround...
from channels import *





# this represents the main window
# and also contains most application behaviour
class StreamTunerTwo(gtk.glade.XML):

        # constructor
        def __init__(self):

            # instantiate gtk/glade widgets in current object
            gtk.glade.XML.__init__(self, conf.share+"/st2.glade")
                # gtk.glade.xml_new_from_buffer...
                
            # non-glade widgets
            self.widgets = {}

            # helper vars
            self.sbar_cid = self.get_widget("statusbar").get_context_id("messages")
            self.sbar_msg = []
            # threads
            self.working = []

            # channel modules
            self.c = {}  # needs to exist before instantiating modules, else any early __getattr__ access on self.c[] creates race condition
            self.c = {
              "bookmarks": bookmarks(parent=self),
              "special": None,
              "shoutcast": shoutcast(parent=self),
              "live365": None,
              "xiph": None,
#              "search": None,
#              "google": None,
            }
            self.cn = ["bookmarks", "special", "shoutcast", "live365", "xiph"]    # order of channel notebook tabs
            self.cc = "shoutcast"    # currently selected channel name (as index in self.c{})
            
            # plugin channel modules
            self.load_plugin_channels()
            

            # set some defaults
            winlayout = conf.load("window")
            if (winlayout):
                mygtk.app_restore(self, winlayout)
            winstate = conf.load("state")
            if (winstate):
                for id in winstate.keys():
                    self.c[id].current = winstate[id]["current"]
                    self.c[id].shown = winlayout[id+"_list"].get("row:selected", 0)

            # display current open channel/notebook tab
            self.c[self.cc].first_show()

      
            # bind gtk/glade event names to functions
            self.signal_autoconnect({
                "gtk_main_quit" : self.gtk_main_quit,                # close window
                # treeviews / notebook
                "on_stream_row_activated" : self.on_play_clicked,    # double click in a streams list
                "on_category_clicked": self.on_category_clicked,     # new selection in category list
                "on_notebook_channels_switch_page": self.channel_switch,   # channel notebook tab changed
                # toolbar
                "on_play_clicked" : self.on_play_clicked,
                "on_record_clicked": self.on_record_clicked,
                "on_homepage_stream_clicked": self.on_homepage_stream_clicked,
                "on_reload_clicked": self.on_reload_clicked,
                "on_stop_clicked": self.on_stop_clicked,
                "on_homepage_channel_clicked" : self.on_homepage_channel_clicked,
                # menu
                "menu_toolbar_standard": lambda(w): (self.toolbar.unset_style(), self.toolbar.unset_icon_size()),
                "menu_toolbar_style_icons": lambda(w): (self.toolbar.set_style(gtk.TOOLBAR_ICONS)),
                "menu_toolbar_style_both": lambda(w): (self.toolbar.set_style(gtk.TOOLBAR_BOTH)),
                "menu_toolbar_size_small": lambda(w): (self.toolbar.set_icon_size(gtk.ICON_SIZE_SMALL_TOOLBAR)),
                "menu_toolbar_size_medium": lambda(w): (self.toolbar.set_icon_size(gtk.ICON_SIZE_DND)),
                "menu_toolbar_size_large": lambda(w): (self.toolbar.set_icon_size(gtk.ICON_SIZE_DIALOG)),
                # else
                "menu_properties": self.config_open,
                "config_cancel": lambda(w): (self.win_config.hide()),
                "config_save": self.config_save,
                "update_categories": self.update_categories,
                "app_state": self.app_state,
                "bookmark": self.bookmark,
                "save_as": self.save_as,
                "menu_about": self.menu_about,
                "menu_copy": self.menu_copy,
            })


            # actually display main window
            self.win_streamtuner2.show()

          

        # shortcut fo glade.get_widget(),
        # allows access to widgets as direct attributes instead of using .get_widget()
        # or looking in self.c[] for the named channel plugins
        def __getattr__(self, name):
            if (self.c.has_key(name)):
                return self.c[name]
            else:
                return self.get_widget(name)
            # or gives an error, if neither exists
        def get_widget(self, name):
            if self.__dict__["widgets"].has_key(name):
                return self.__dict__["widgets"][name]
            else:
                return gtk.glade.XML.get_widget(self, name)


        # play button
        def on_play_clicked(self, widget, event=None, *more):
            url = self.selected(0, "url");
            #audioformat = self.selected(0, "format");
            action.play(url, audioformat="audio/mp3", listformat=self.c[self.cc].listformat)

        # streamripper
        def on_record_clicked(self, widget):
            url = self.selected(0, "url")
            action.record(url, "audio/mp3")

        # browse stream
        def on_homepage_stream_clicked(self, widget):
            url = self.selected(0, "homepage")             
            action.browser(url)
             
        # browse channel
        def on_homepage_channel_clicked(self, widget):
            action.browser(self.c[self.cc].homepage)
            

        # reload stream list in current channel-category
        def on_reload_clicked(self, widget=None, reload=1):
            thread = Thread(target=self.c[self.cc].load, args=(self.c[self.cc].current, reload,))
            thread.start()
            self.working.append(thread)
        # stop
        def on_stop_clicked(self, widget):
            if self.working:
                thread = self.working.pop()
                thread.stop()

        # doubleclick in stream list
        #def on_shoutcast_list_row_activated(self, widget, event, *more):
        #    self.on_play_clicked(widget)
        
        # click in category list
        def on_category_clicked(self, widget, event, *more):
            category = self.c[self.cc].currentcat()
            self.on_reload_clicked(None, reload=0)
            pass

        # fetches varname from treeview selection            
        def selected(self, widget, name="url"):
            if not widget:
                widget = self.c[self.cc].gtk_list
            (model, iter) = widget.get_selection().get_selected()
            if (model and iter):
                liststore_index = self.c[self.cc].rowmap.index(name)
                return model.get_value(iter, liststore_index)
            else:
                pass
                
        # shortcut to statusbar
        # (hacked to work from within threads, circumvents the statusbar msg pool actually)
        def status(self, text=""):
            # remove text
            while ((not text) and (type(text)==str) and len(self.sbar_msg)):
                self.sbar_msg.pop()
                gobject.idle_add(lambda:self.statusbar.pop(self.sbar_cid) and False)
            # progressbar
            if (type(text)==float):
                if (text >= 1.0):  # completed
                    gobject.idle_add(lambda:self.progress.hide() and False)
                else:  # show percentage
                    gobject.idle_add(lambda:self.progress.show() or self.progress.set_fraction(text) and False)
                    if (text <= 0.0):  # unknown state
                        gobject.idle_add(lambda:self.progress.pulse() and False)
            # add text
            elif (type(text)==str):
                self.sbar_msg.append(1)
                gobject.idle_add(lambda:self.statusbar.push(self.sbar_cid, text) and False)
            pass

        # end application and gtk+ main loop
        def gtk_main_quit(self, widget):
            gtk.main_quit()

        # store window/widget states (sizes, selections, etc.)
        def app_state(self, widget):
            # gtk widget states
            widgetnames = ["win_streamtuner2", "toolbar", "notebook_channels", ] \
                        + [id+"_list" for id in self.cn] + [id+"_cat" for id in self.cn]
            conf.save("window", mygtk.app_state(wTree=self, widgetnames=widgetnames))
            # object vars
            channelopts = {} #dict([(id, {"current":self.c[id].current}) for id in self.cn])
            for id in self.c.keys():
                if (self.c[id]):
                    channelopts[id] = {"current":self.c[id].current}
            conf.save("state", channelopts)

        # notebook tab clicked
        def channel_switch(self, notebook, page, page_num, *args):
            self.cc = self.cn[page_num]
            #-- if first selected, load current cagtegory
            self.c[self.cc].first_show()

        # complete dict of currently selected entry
        # - see also self.selected()
        def row(self):
            (model, iter) = self.c[self.cc].gtk_list.get_selection().get_selected()
            row = {}
            for i,id in enumerate(self.c[self.cc].rowmap):
                val = model.get_value(iter, i)
                row[id] = val
            return row        
            
        # add current selection to bookmark store
        def bookmark(self, widget):
            self.bookmarks.add(row)
            # code to update current list (set icon just in on-screen liststore, it would be updated with next display() anyhow - and there's no need to invalidate the ls cache, because that's referenced by model anyhow)
            model.set_value(iter, 0, gtk.STOCK_ABOUT)
            # refresh bookmarks tab
            self.bookmarks.load(self.bookmarks.default)

        # display win_config, pre-fill text fields from global conf. object
        def config_open(self, widget):
            self.config_apply(conf.__dict__, "config_", 0)
            #self.win_config.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse('#997777'))
            self.win_config.show()
        
        # set/load values between gtk window and conf. dict
        def config_apply(self, config, prefix="config_", save=0):
            for key,val in config.iteritems():
                # map non-alphanumeric chars from config{} to underscores in according gtk widget names
                id = re.sub("[^\w]", "_", key)
                w = self.get_widget(prefix + id)
           #print save, prefix+id, w, val
                # recurse into dictionaries
                if (type(val) == dict):
                    self.config_apply(val, prefix + id + "_", save)
                # load or set gtk.Entry text field
                elif (w and save and type(w)==gtk.Entry):
                    config[key] = w.get_text()
                elif (w and type(w)==gtk.Entry):
                    w.set_text(val)
                elif (w and save):
                    config[key] = w.get_active()
                elif (w):
                    w.set_active(val)
            pass
        
        # save config
        def config_save(self, widget):
            self.config_apply(conf.__dict__, "config_", 1)
            conf.save()
            self.win_config.hide()
            
        # reload category tree
        def update_categories(self, widget):
            c = self.c[self.cc]
            # load from network, and save title list in cache/ file
            c.update_categories()
            conf.save("cache/categories_"+c.module, c.categories)
            # rebuild gtk.TreeView
            c.gtk_cat.remove_column(c.gtk_cat.get_column(0))
            mygtk.tree(c.gtk_cat, c.categories, title="Category", icon=gtk.STOCK_OPEN);

        # save a file            
        def save_as(self, widget):
            row = self.row()
            default_fn = row["title"] + ".m3u"
            fn = mygtk.save_file("Save Stream", None, default_fn, [(".m3u","*m3u"),(".pls","*pls"),(".xspf","*xspf"),(".smil","*smil"),(".asx","*asx"),("all files","*")])
            if fn:
                action.save(row, fn)
            pass
            
        # about us
        def menu_about(self, widget):
            a = gtk.AboutDialog()
            a.set_version("1.9.8")
            a.set_name("streamtuner2")
            a.set_license("Public Domain\n\nNo Strings Attached.\nUnrestricted distribution,\nmodification, use.")
            a.set_authors(["mario salzer <xmilky+st2@gm..>\n"])
            a.set_website("http://milki.erphesfurt.de/streamtuner2/")
            a.connect("response", lambda widget, ok: ( widget.hide(),widget.destroy() ) )
            a.show()

        # save current stream URL into clipboard
        def menu_copy(self, w):
            gtk.clipboard_get().set_text(self.selected(0, "url"))



        # load plugins from /usr/share/streamtuner2/channels dir
        def load_plugin_channels(self):
             ###   sys.path.append(conf.dir + "/channels/")
            # find plugin files
            for fn in os.listdir(conf.share + "/channels/"):
                if (re.match("^[a-z0-9]+\.py$", fn)):
                    module = fn[0:-3] # cut filename
                    
                    # load plugin
                    plugin = __import__("channels."+module, None, None, [""])
                    plugin_class = plugin.__dict__[module]

                    # add and initialize object
                    if module not in self.cn:
                        self.cn.append(module)
                    self.c[module] = plugin_class(parent=self)
            pass
                










#-- favourite lists                                            ------------------------------------------
# this module lists static content from ~/.config/streamtuner2/bookmarks.json only,
# its data list is queries by other plugins to add 'star' icons
class bookmarks(GenericChannel):

        # desc
        api = "st2"
        module = "bookmarks"
        version = 0.3
        base_url = "file:.config/streamtuner2/bookmarks.json"
        listformat = "*/*"

        # content
        categories = ["favourite",]
        current = "favourite"
        default = "favourite"
        streams = {"favourite":[]}
        urls = []


        # this channel does not actually load any data
        def update_categories(self):
            pass
        def update_streams(self):
            pass
        # except from the cache file
        def cache(self):
            # stream list
            cache = conf.load(self.module)  # becomes "bookmarks.json"
            if (cache):
                self.streams = cache
            pass
        # save to cache file
        def save(self):
            conf.save(self.module, self.streams)
            pass


        # checks for existence of an URL in bookmarks store,
        # this method is called by other channel modules' display() method
        def is_in(self, url, once=1):
            if (not self.urls):
                self.urls = [row.get("url","urn:x-streamtuner2:no") for row in self.streams["favourite"]]
            return url in self.urls

        # called from main window / menu / context menu,
        # when bookmark is to be added for a selected stream entry
        def add(self, row):
            self.streams["favourite"].append(row)
            row["favourite"] = 1  # send status back into source streams list
            self.urls.append(row["url"])
            self.save()
            self.load(self.default)
            pass

        # simplified gtk TreeStore display logic (just one category for the moment, always rebuilt)
        def load(self, category, force=False):
            #self.liststore[category] = \
            mygtk.columns(self.gtk_list, self.datamap, self.display(self.streams[category]))
            pass












# SHOUTcast data module                                          ----------------------------------------
class shoutcast(GenericChannel):

        # desc
        api = "st2"
        module = "shoutcast"
        version = 0.8
        homepage = "http://www.shoutcast.com/" 
        base_url = "http://classic.shoutcast.com/"
        listformat = "audio/x-scpls"

        # categories
        categories = ["default", [], 'TopTen', [], 'Alternative', ['College', 'Emo', 'Hardcore', 'Industrial', 'Punk', 'Ska'], 'Americana', ['Bluegrass', 'Blues', 'Cajun', 'Folk'], 'Classical', ['Contemporary', 'Opera', 'Symphonic'], 'Country', ['Bluegrass', 'New Country', 'Western Swing'], 'Electronic', ['Acid Jazz', 'Ambient', 'Breakbeat', 'Downtempo', 'Drum and Bass', 'House', 'Trance', 'Techno'], 'Hip Hop', ['Alternative', 'Hardcore', 'New School', 'Old School', 'Turntablism'], 'Jazz', ['Acid Jazz', 'Big Band', 'Classic', 'Latin', 'Smooth', 'Swing'], 'Pop/Rock', ['70s', '80s', 'Classic', 'Metal', 'Oldies', 'Pop', 'Rock', 'Top 40'], 'R&B/Soul', ['Classic', 'Contemporary', 'Funk', 'Smooth', 'Urban'], 'Spiritual', ['Alternative', 'Country', 'Gospel', 'Pop', 'Rock'], 'Spoken', ['Comedy', 'Spoken Word', 'Talk'], 'World', ['African', 'Asian', 'European', 'Latin', 'Middle Eastern', 'Reggae'], 'Other/Mixed', ['Eclectic', 'Film', 'Instrumental']]
        current = ""
        default = "default"
        empty = "default"
        

            
        # extracts the category list from shoutcast.com,
        # the text/html format always being:
        #        <OPTION VALUE="Classical">Classical
        #                <OPTION VALUE="Contemporary"> - Contemporary
        #                <OPTION VALUE="Opera"> - Opera
        # sub-categories are indented by two tabs, maincats just by one
        def update_categories(self):
            html = http.get(self.base_url + "")
            self.categories = ["default"]
            sub = []
            for uu in re.findall(r'<OPTION VALUE="(.+?)">( - )?(.+)', html):
                name = urllib.unquote(uu[0])
                # main category
                if (len(uu[1]) == 0):
                    self.categories.append(name)
                    sub = []
                    self.categories.append(sub)
                # sub entry
                else:
                    sub.append(name)
            # it's done
            #print self.categories
            conf.save("cache/categories_shoutcast", self.categories)
            pass



        # downloads stream list from shoutcast for given category
        def update_streams(self, cat, search=""):
            if (not cat or cat == self.empty):
                return []
        
            # regular patterns
            rx_stream = re.compile("""
                 href="/(sbin/shoutcast-playlist.pls[^"]+)">.+?
                 size="1"><b>\[([^\]]+?)\]</.+?
                 <a.id="listlinks".target="_scurl".href="([^"]+)">([^<>]+)</a>.+?
                 >Now.Playing:</font>\s*([^<>]+)</font>.+?
                 (\d+)/(\d+).+?
                 >(\d+)</font>.+?
                 <!--.color.AAC\+.Type.data.-->.+?>\s*([\w\d]+)[+]*<
                 """, re.S|re.I|re.X)
            rx_next = re.compile(">\[ Next \]</a>")
            
            # loop
            entries = []
            next = 0
            while (next<2000):
            
               # get page
               url = self.base_url + "directory/index.phtml?sgenre="+urllib.quote(cat)+"&maxbitrate=999&numresult=100&startat="+str(next)+"&s="+search
               #print url
               html = http.get(url, feedback=self.parent.status)
               #print "rx"              
               
               # extract entries
               self.parent.status("parsing document...")
               for uu in rx_stream.findall(html):
                   #print uu
                   entries += [{
                       "url": self.base_url + uu[0],
                       "genre": uu[1],
                       "homepage": http.fix_url(uu[2]),
                       "title": entity_decode(uu[3]),
                       "playing": entity_decode(uu[4]),
                       "listeners": int(uu[5]),
                       "max": int(uu[6]),
                       "bitrate": int(uu[7]),
                       "format": "audio/" + uu[8].lower(),
                   }]
               self.parent.status()
               
               # more pages to load?
               if (re.search(rx_next, html)):
                  next += 100
               else:
                  next = 99999
            
            #fin
            return entries

















#-- run main                                ---------------------------------------------
if __name__ == "__main__":

    #-- global configuration settings
    # >conf = Config()< already happened with >from config import conf<
    
    # prepare for threading in Gtk+ callbacks
    gobject.threads_init()
    #gtk.gdk.threads_init()  # for _enter/leave locking method
    
    # prepare main window
    main = StreamTunerTwo()
    action.main = main
    action = action.action
    
    # first invocation
    if (conf.__dict__.get("firstrun")):
        main.config_open(None)
        del conf.firstrun

    # run
    gtk.main()



#
#
#
#